# SPDX-FileCopyrightText: 2024 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import pickle

import lox_space as lox
import pytest


@pytest.mark.parametrize(
    "obj",
    [
        lox.Origin("Earth"),
        lox.Frame("ICRF"),
        lox.ElevationMask.fixed(0.0 * lox.rad),
        lox.TimeScale("TAI"),
        lox.Time("TAI", 2000, 1, 1),
    ],
)
def test_pickle(obj):
    pickled = pickle.dumps(obj)
    unpickled = pickle.loads(pickled)
    assert unpickled == obj



# SPDX-FileCopyrightText: 2025 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

"""
OneWeb Europe Coverage Benchmark

This benchmark tests visibility analysis performance with a realistic large-scale scenario:
- 651 OneWeb satellites from TLE data (using existing oneweb fixture)
- Ground points covering Europe with 1-degree resolution (~630 points)
- Total combinations: ~410,000 visibility calculations

This represents a realistic worst-case scenario for satellite visibility analysis.
"""

import time
import pytest
import numpy as np
import lox_space as lox


def create_europe_ground_assets(resolution_deg=1.0, min_elevation_deg=10.0):
    """
    Create a list of ground assets covering Europe with specified resolution.

    Parameters
    ----------
    resolution_deg : float
        Grid resolution in degrees (default: 1.0)
    min_elevation_deg : float
        Minimum elevation angle in degrees (default: 10.0)

    Returns
    -------
    list[lox.GroundStation]
        List of ground assets covering Europe.
    """
    # Europe bounding box (approximate)
    lon_min, lon_max = -10.0, 40.0  # West to East
    lat_min, lat_max = 35.0, 70.0  # South to North

    # Generate grid points
    lons = np.arange(lon_min, lon_max + resolution_deg, resolution_deg)
    lats = np.arange(lat_min, lat_max + resolution_deg, resolution_deg)

    elevation_mask = lox.ElevationMask.fixed(min_elevation_deg * lox.deg)
    origin = lox.Origin("Earth")

    ground_assets = []
    point_count = 0
    for lon in lons:
        for lat in lats:
            # Filter to approximate Europe shape (rough filtering)
            if is_point_in_europe(lon, lat):
                gs_name = f"EU_GS_{point_count:04d}"
                ground_location = lox.GroundLocation(
                    origin=origin,
                    longitude=lon * lox.deg,
                    latitude=lat * lox.deg,
                    altitude=0.0 * lox.km,  # Sea level
                )
                ground_assets.append(
                    lox.GroundStation(gs_name, ground_location, elevation_mask)
                )
                point_count += 1

    return ground_assets


def is_point_in_europe(lon, lat):
    """
    Rough filter to include only points that are approximately in Europe.
    This is a simplified polygon check for demonstration purposes.
    """
    # Very rough Europe boundary - exclude ocean areas
    if lat < 36.0:  # Below southern Europe
        return False
    if lat > 71.0:  # Above northern Europe
        return False
    if lon < -9.0:  # West of Ireland
        return False
    if lon > 40.0:  # East of Urals
        return False

    # Exclude some obvious non-Europe areas
    if lon < -5.0 and lat < 43.0:  # Atlantic Ocean
        return False
    if lon > 30.0 and lat < 45.0:  # Turkey/Middle East area
        return False
    if lon > 35.0 and lat > 60.0:  # Far eastern Russia
        return False

    return True


@pytest.fixture(scope="session")
def europe_ground_assets():
    """Fixture providing Europe ground asset grid."""
    return create_europe_ground_assets(resolution_deg=1.0)


@pytest.fixture(scope="session")
def europe_ground_assets_coarse():
    """Fixture providing coarser Europe ground asset grid for faster testing."""
    return create_europe_ground_assets(resolution_deg=2.0)


@pytest.fixture(scope="session")
def t0(oneweb):
    return next(iter(oneweb.values())).time()


@pytest.fixture(scope="session")
def t1(t0):
    return t0 + lox.TimeDelta(7200)


def make_space_assets(oneweb_dict):
    return [lox.Spacecraft(name, sgp4) for name, sgp4 in oneweb_dict.items()]


def make_scenario_and_ensemble(oneweb_dict, ground_assets, t0, t1):
    space_assets = make_space_assets(oneweb_dict)
    scenario = lox.Scenario(
        t0, t1, spacecraft=space_assets, ground_stations=ground_assets
    )
    ensemble = scenario.propagate()
    return scenario, ensemble


@pytest.fixture(scope="session")
def oneweb_sample_small(oneweb):
    """Small sample of OneWeb constellation for quick tests."""
    return dict(list(oneweb.items())[:10])


@pytest.fixture(scope="session")
def oneweb_sample_medium(oneweb):
    """Medium sample of OneWeb constellation for performance tests."""
    return dict(list(oneweb.items())[:50])


@pytest.fixture(scope="session")
def oneweb_sample_large(oneweb):
    """Large sample of OneWeb constellation for stress tests."""
    return dict(list(oneweb.items())[:200])


@pytest.fixture(scope="session")
def small_scenario_and_ensemble(oneweb_sample_small, europe_ground_assets_coarse, t0, t1):
    """Pre-computed scenario and ensemble for small benchmark."""
    sample_ground_assets = europe_ground_assets_coarse[:10]
    return make_scenario_and_ensemble(oneweb_sample_small, sample_ground_assets, t0, t1)


@pytest.fixture(scope="session")
def medium_scenario_and_ensemble(oneweb_sample_medium, europe_ground_assets_coarse, t0, t1):
    """Pre-computed scenario and ensemble for medium benchmark."""
    sample_ground_assets = europe_ground_assets_coarse[:50]
    return make_scenario_and_ensemble(oneweb_sample_medium, sample_ground_assets, t0, t1)


@pytest.fixture(scope="session")
def large_scenario_and_ensemble(oneweb_sample_large, europe_ground_assets, t0, t1):
    """Pre-computed scenario and ensemble for large benchmark."""
    sample_ground_assets = europe_ground_assets[:100]
    return make_scenario_and_ensemble(oneweb_sample_large, sample_ground_assets, t0, t1)


class TestOneWebEuropeBenchmark:
    """Benchmark tests for OneWeb Europe coverage scenario."""

    def test_benchmark_info(self, europe_ground_assets, oneweb):
        """Display benchmark scenario information."""
        num_spacecraft = len(oneweb)
        num_ground_stations = len(europe_ground_assets)
        total_combinations = num_spacecraft * num_ground_stations

        print(f"\n{'=' * 60}")
        print("OneWeb Europe Coverage Benchmark")
        print(f"{'=' * 60}")
        print(f"Spacecraft (OneWeb): {num_spacecraft}")
        print(f"Ground stations (Europe 1deg): {num_ground_stations}")
        print(f"Total combinations: {total_combinations:,}")
        print(f"Expected visibility calculations: {total_combinations:,}")
        print(f"{'=' * 60}")

        # This is not a real test, just info display
        assert num_spacecraft > 600  # Should have most OneWeb satellites
        assert num_ground_stations > 500  # Should cover Europe well
        assert total_combinations > 300000  # Should be a large-scale test

    def test_benchmark_sequential_sample(
        self,
        small_scenario_and_ensemble,
        oneweb_sample_small,
        europe_ground_assets_coarse,
        ephemeris,
    ):
        """Test visibility analysis on a small sample."""
        scenario, ensemble = small_scenario_and_ensemble
        analysis = lox.VisibilityAnalysis(scenario, ensemble=ensemble)
        results = analysis.compute(ephemeris)

        sample_ground_assets = europe_ground_assets_coarse[:10]
        assert results.num_pairs() == len(oneweb_sample_small) * len(
            sample_ground_assets
        )

    @pytest.mark.benchmark
    def test_benchmark_parallel_medium(
        self,
        medium_scenario_and_ensemble,
        oneweb_sample_medium,
        europe_ground_assets_coarse,
        ephemeris,
    ):
        """Test visibility analysis on a medium sample."""
        scenario, ensemble = medium_scenario_and_ensemble
        analysis = lox.VisibilityAnalysis(scenario, ensemble=ensemble)
        results = analysis.compute(ephemeris)

        sample_ground_assets = europe_ground_assets_coarse[:50]
        assert results.num_pairs() == len(oneweb_sample_medium) * len(
            sample_ground_assets
        )

    @pytest.mark.slow
    def test_benchmark_parallel_large(
        self,
        large_scenario_and_ensemble,
        oneweb_sample_large,
        europe_ground_assets,
        ephemeris,
    ):
        """Test visibility analysis on a large subset (marked as slow)."""
        scenario, ensemble = large_scenario_and_ensemble
        analysis = lox.VisibilityAnalysis(scenario, ensemble=ensemble)
        results = analysis.compute(ephemeris)

        sample_ground_assets = europe_ground_assets[:100]
        assert results.num_pairs() == len(oneweb_sample_large) * len(
            sample_ground_assets
        )

    @pytest.mark.slow
    @pytest.mark.benchmark
    def test_full_scale_estimate(self, europe_ground_assets, oneweb):
        """Estimate performance for full-scale scenario."""
        total_spacecraft = len(oneweb)
        total_ground_stations = len(europe_ground_assets)
        total_combinations = total_spacecraft * total_ground_stations

        print(f"\n{'=' * 60}")
        print("Full-Scale Performance Estimate")
        print(f"{'=' * 60}")
        print(f"Total OneWeb satellites: {total_spacecraft}")
        print(f"Total Europe ground points: {total_ground_stations}")
        print(f"Total combinations: {total_combinations:,}")
        print(
            f"Memory estimate (results only): {total_combinations * 200 / 1024 / 1024:.1f} MB"
        )
        print(f"{'=' * 60}")

        # This is just an informational test
        assert total_combinations > 400000



# SPDX-FileCopyrightText: 2026 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import math
import pickle

import pytest
import lox_space as lox


# --- Decibel ---


def test_decibel_new():
    db = lox.Decibel(3.0)
    assert float(db) == 3.0


def test_decibel_from_linear():
    db = lox.Decibel.from_linear(100.0)
    assert float(db) == pytest.approx(20.0, abs=1e-10)


def test_decibel_to_linear():
    db = lox.Decibel(20.0)
    assert db.to_linear() == pytest.approx(100.0, abs=1e-10)


def test_decibel_roundtrip():
    db = lox.Decibel(13.5)
    assert float(lox.Decibel.from_linear(db.to_linear())) == pytest.approx(
        13.5, abs=1e-10
    )


def test_decibel_add():
    a = lox.Decibel(3.0)
    b = lox.Decibel(3.0)
    assert float(a + b) == pytest.approx(6.0, abs=1e-10)


def test_decibel_sub():
    a = lox.Decibel(6.0)
    b = lox.Decibel(3.0)
    assert float(a - b) == pytest.approx(3.0, abs=1e-10)


def test_decibel_neg():
    db = lox.Decibel(3.0)
    assert float(-db) == pytest.approx(-3.0, abs=1e-10)


def test_decibel_repr():
    db = lox.Decibel(3.0)
    assert repr(db) == "Decibel(3.0)"


def test_decibel_str():
    db = lox.Decibel(3.0)
    assert str(db) == "3 dB"


def test_decibel_eq():
    assert lox.Decibel(3.0) == lox.Decibel(3.0)
    assert not (lox.Decibel(3.0) == lox.Decibel(4.0))


def test_decibel_pickle():
    db = lox.Decibel(13.5)
    assert pickle.loads(pickle.dumps(db)) == db


def test_decibel_repr_roundtrip():
    db = lox.Decibel(13.5)
    assert eval(repr(db), {"Decibel": lox.Decibel}) == db


def test_decibel_mul():
    db = 14.0 * lox.dB
    assert float(db) == pytest.approx(14.0, abs=1e-10)


def test_decibel_rmul():
    db = lox.dB * 14.0
    assert float(db) == pytest.approx(14.0, abs=1e-10)


# --- Modulation ---


def test_modulation_bits_per_symbol():
    assert lox.Modulation("BPSK").bits_per_symbol() == 1
    assert lox.Modulation("QPSK").bits_per_symbol() == 2
    assert lox.Modulation("8PSK").bits_per_symbol() == 3
    assert lox.Modulation("16QAM").bits_per_symbol() == 4
    assert lox.Modulation("256QAM").bits_per_symbol() == 8


def test_modulation_invalid():
    with pytest.raises(ValueError, match="unknown modulation"):
        lox.Modulation("AM")


def test_modulation_repr():
    m = lox.Modulation("QPSK")
    assert repr(m) == "Modulation('QPSK')"


def test_modulation_eq():
    assert lox.Modulation("BPSK") == lox.Modulation("BPSK")
    assert not (lox.Modulation("BPSK") == lox.Modulation("QPSK"))


def test_modulation_pickle():
    for name in ["BPSK", "QPSK", "8PSK", "16QAM", "32QAM", "64QAM", "128QAM", "256QAM"]:
        m = lox.Modulation(name)
        assert pickle.loads(pickle.dumps(m)) == m


def test_modulation_repr_roundtrip():
    m = lox.Modulation("8PSK")
    assert eval(repr(m), {"Modulation": lox.Modulation}) == m


# --- Parabolic Pattern ---


def test_parabolic_peak_gain():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    gain = p.peak_gain(frequency=29e9 * lox.Hz)
    assert float(gain) == pytest.approx(46.01119, rel=1e-4)


def test_parabolic_beamwidth():
    # D=0.98m, f=29GHz: full HPBW = 2·arcsin(1.6163308·λ/(π·D)) ≈ 0.6219°
    # (the previous value 0.7372° was the first-null angle, not HPBW)
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    bw = p.beamwidth(frequency=29e9 * lox.Hz)
    lam = 299_792_458 / 29e9
    expected = math.degrees(2.0 * math.asin(1.6163308 * lam / (math.pi * 0.98)))
    assert bw.to_degrees() == pytest.approx(expected, rel=1e-4)


def test_parabolic_beamwidth_none_for_sub_wavelength_diameter():
    # At 1 GHz (λ ≈ 0.300 m) the HPBW threshold diameter is ≈ 0.154 m.
    # A 0.1 m dish is below this limit, so beamwidth is undefined.
    p = lox.ParabolicPattern(diameter=0.1 * lox.m, efficiency=0.65)
    assert p.beamwidth(frequency=1e9 * lox.Hz) is None


def test_parabolic_on_axis_equals_peak():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    f = 29e9 * lox.Hz
    gain = p.gain(f, angle=0.0 * lox.deg)
    peak = p.peak_gain(f)
    assert float(gain) == pytest.approx(float(peak), abs=1e-6)


def test_parabolic_gain_at_180():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    gain = p.gain(29e9 * lox.Hz, angle=180.0 * lox.deg)
    assert float(gain) < -50.0


def test_parabolic_from_beamwidth_roundtrip():
    p = lox.ParabolicPattern.from_beamwidth(
        beamwidth=math.degrees(0.1) * lox.deg, frequency=2e9 * lox.Hz, efficiency=0.65
    )
    bw = p.beamwidth(frequency=2e9 * lox.Hz)
    assert bw.to_degrees() == pytest.approx(math.degrees(0.1), rel=0.01)


def test_parabolic_eq():
    a = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    b = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    c = lox.ParabolicPattern(diameter=1.0 * lox.m, efficiency=0.45)
    assert a == b
    assert not (a == c)


def test_parabolic_pickle():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    assert pickle.loads(pickle.dumps(p)) == p


def test_parabolic_repr_roundtrip():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    assert (
        eval(
            repr(p),
            {"ParabolicPattern": lox.ParabolicPattern, "Distance": lox.Distance},
        )
        == p
    )


# --- Gaussian Pattern ---


def test_gaussian_peak_gain():
    p = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    gain = p.peak_gain(frequency=29e9 * lox.Hz)
    assert float(gain) == pytest.approx(46.01119, rel=1e-4)


def test_gaussian_3db_at_half_beamwidth():
    p = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    f = 29e9 * lox.Hz
    bw = p.beamwidth(f)
    half_bw = bw * 0.5
    peak = float(p.peak_gain(f))
    gain = float(p.gain(f, angle=half_bw))
    assert peak - gain == pytest.approx(3.0103, abs=0.01)


def test_gaussian_eq():
    a = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    b = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    c = lox.GaussianPattern(diameter=1.0 * lox.m, efficiency=0.45)
    assert a == b
    assert not (a == c)


def test_gaussian_pickle():
    p = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    assert pickle.loads(pickle.dumps(p)) == p


def test_gaussian_repr_roundtrip():
    p = lox.GaussianPattern(diameter=0.98 * lox.m, efficiency=0.45)
    assert (
        eval(
            repr(p), {"GaussianPattern": lox.GaussianPattern, "Distance": lox.Distance}
        )
        == p
    )


# --- Dipole Pattern ---


def test_half_wave_dipole_broadside():
    # Half-wave dipole: peak gain at 90 deg approx 2.15 dBi
    c = 299792458.0
    wavelength = c / 29e9
    d = lox.DipolePattern(length=(wavelength / 2.0) * lox.m)
    gain = d.gain(29e9 * lox.Hz, angle=90.0 * lox.deg)
    assert float(gain) == pytest.approx(2.15, abs=0.01)


def test_half_wave_dipole_endfire():
    c = 299792458.0
    wavelength = c / 29e9
    d = lox.DipolePattern(length=(wavelength / 2.0) * lox.m)
    gain = d.gain(29e9 * lox.Hz, angle=0.0 * lox.deg)
    assert float(gain) < -50.0


def test_short_dipole_peak():
    c = 299792458.0
    wavelength = c / 29e9
    d = lox.DipolePattern(length=(wavelength / 100.0) * lox.m)
    peak = d.peak_gain(29e9 * lox.Hz)
    assert float(peak) == pytest.approx(1.76, abs=0.1)


def test_dipole_eq():
    a = lox.DipolePattern(length=0.5 * lox.m)
    b = lox.DipolePattern(length=0.5 * lox.m)
    c = lox.DipolePattern(length=1.0 * lox.m)
    assert a == b
    assert not (a == c)


def test_dipole_pickle():
    d = lox.DipolePattern(length=0.5 * lox.m)
    assert pickle.loads(pickle.dumps(d)) == d


def test_dipole_repr_roundtrip():
    d = lox.DipolePattern(length=0.5 * lox.m)
    assert (
        eval(repr(d), {"DipolePattern": lox.DipolePattern, "Distance": lox.Distance})
        == d
    )


def test_complex_antenna_dipole_beamwidth_is_none():
    # DipolePattern returns None because the API has no plane-of-measurement
    # parameter; the E-plane HPBW is physically defined but not expressible as
    # a single value here. ComplexAntenna delegates to the pattern, so it also
    # returns None.
    d = lox.DipolePattern(length=0.005 * lox.m)
    a = lox.ComplexAntenna(pattern=d, boresight=[0.0, 0.0, 1.0])
    assert a.beamwidth(frequency=29e9 * lox.Hz) is None


# --- Antennas ---


def test_simple_antenna():
    a = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    r = repr(a)
    assert r.startswith("SimpleAntenna(gain=Decibel(30.0), beamwidth=Angle(")


def test_simple_antenna_eq():
    a = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    b = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    c = lox.SimpleAntenna(gain=20.0 * lox.dB, beamwidth=3.0 * lox.deg)
    assert a == b
    assert not (a == c)


def test_simple_antenna_pickle():
    a = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    assert pickle.loads(pickle.dumps(a)) == a


def test_simple_antenna_repr_roundtrip():
    a = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    assert (
        eval(
            repr(a),
            {
                "SimpleAntenna": lox.SimpleAntenna,
                "Decibel": lox.Decibel,
                "Angle": lox.Angle,
            },
        )
        == a
    )


def test_complex_antenna():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    a = lox.ComplexAntenna(pattern=p, boresight=[0.0, 0.0, 1.0])
    gain = a.gain(29e9 * lox.Hz, angle=0.0 * lox.deg)
    assert float(gain) == pytest.approx(46.01119, rel=1e-4)


def test_complex_antenna_invalid_pattern():
    with pytest.raises(ValueError, match="expected a ParabolicPattern"):
        lox.ComplexAntenna(pattern="not a pattern", boresight=[0.0, 0.0, 1.0])


def test_complex_antenna_pickle():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    a = lox.ComplexAntenna(pattern=p, boresight=[0.0, 0.0, 1.0])
    restored = pickle.loads(pickle.dumps(a))
    # ComplexAntenna doesn't have __eq__, verify via gain
    assert float(restored.gain(29e9 * lox.Hz, 0.0 * lox.deg)) == pytest.approx(
        float(a.gain(29e9 * lox.Hz, 0.0 * lox.deg)), abs=1e-10
    )


def test_complex_antenna_repr():
    p = lox.ParabolicPattern(diameter=0.98 * lox.m, efficiency=0.45)
    a = lox.ComplexAntenna(pattern=p, boresight=[0.0, 0.0, 1.0])
    r = repr(a)
    assert r.startswith("ComplexAntenna(pattern=ParabolicPattern(")
    assert "boresight=[0.0, 0.0, 1.0]" in r


# --- Transmitter ---


def test_transmitter_eirp():
    # 10 dBi antenna, 5 W power, 1 dB line loss, 0 dB OBO
    # EIRP = 10 + 10*log10(5) - 1 = 15.99 dBW
    a = lox.SimpleAntenna(gain=10.0 * lox.dB, beamwidth=10.0 * lox.deg)
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=5.0 * lox.W, line_loss=1.0 * lox.dB
    )
    eirp = tx.eirp(a, angle=0.0 * lox.deg)
    assert float(eirp) == pytest.approx(15.99, abs=0.01)


def test_transmitter_default_obo():
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=0.0 * lox.dB
    )
    assert repr(tx).endswith("output_back_off=Decibel(0.0))")


def test_transmitter_eq():
    a = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    b = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    c = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=5.0 * lox.W, line_loss=1.0 * lox.dB
    )
    assert a == b
    assert not (a == c)


def test_transmitter_pickle():
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz,
        power=10.0 * lox.W,
        line_loss=1.0 * lox.dB,
        output_back_off=0.5 * lox.dB,
    )
    assert pickle.loads(pickle.dumps(tx)) == tx


def test_transmitter_repr_roundtrip():
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz,
        power=10.0 * lox.W,
        line_loss=1.0 * lox.dB,
        output_back_off=0.5 * lox.dB,
    )
    assert (
        eval(
            repr(tx),
            {
                "Transmitter": lox.Transmitter,
                "Frequency": lox.Frequency,
                "Power": lox.Power,
                "Decibel": lox.Decibel,
            },
        )
        == tx
    )


# --- Receivers ---


def test_complex_receiver_noise_temperature():
    rx = lox.ComplexReceiver(
        frequency=29e9 * lox.Hz,
        antenna_noise_temperature=265.0 * lox.K,
        lna_gain=30.0 * lox.dB,
        lna_noise_figure=1.0 * lox.dB,
        noise_figure=5.0 * lox.dB,
        loss=3.0 * lox.dB,
    )
    assert rx.noise_temperature().to_kelvin() == pytest.approx(627.06, rel=1e-4)


def test_complex_receiver_system_noise_temperature():
    rx = lox.ComplexReceiver(
        frequency=29e9 * lox.Hz,
        antenna_noise_temperature=265.0 * lox.K,
        lna_gain=30.0 * lox.dB,
        lna_noise_figure=1.0 * lox.dB,
        noise_figure=5.0 * lox.dB,
        loss=3.0 * lox.dB,
    )
    assert rx.system_noise_temperature().to_kelvin() == pytest.approx(904.53, rel=1e-4)


def test_simple_receiver_eq():
    a = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    b = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    c = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=600.0 * lox.K
    )
    assert a == b
    assert not (a == c)


def test_simple_receiver_pickle():
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    assert pickle.loads(pickle.dumps(rx)) == rx


def test_simple_receiver_repr_roundtrip():
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    assert (
        eval(
            repr(rx),
            {
                "SimpleReceiver": lox.SimpleReceiver,
                "Frequency": lox.Frequency,
                "Temperature": lox.Temperature,
            },
        )
        == rx
    )


def test_complex_receiver_eq():
    kwargs = dict(
        frequency=29e9 * lox.Hz,
        antenna_noise_temperature=265.0 * lox.K,
        lna_gain=30.0 * lox.dB,
        lna_noise_figure=1.0 * lox.dB,
        noise_figure=5.0 * lox.dB,
        loss=3.0 * lox.dB,
    )
    a = lox.ComplexReceiver(**kwargs)
    b = lox.ComplexReceiver(**kwargs)
    c = lox.ComplexReceiver(**{**kwargs, "loss": 4.0 * lox.dB})
    assert a == b
    assert not (a == c)


def test_complex_receiver_pickle():
    rx = lox.ComplexReceiver(
        frequency=29e9 * lox.Hz,
        antenna_noise_temperature=265.0 * lox.K,
        lna_gain=30.0 * lox.dB,
        lna_noise_figure=1.0 * lox.dB,
        noise_figure=5.0 * lox.dB,
        loss=3.0 * lox.dB,
        demodulator_loss=0.5 * lox.dB,
        implementation_loss=0.3 * lox.dB,
    )
    assert pickle.loads(pickle.dumps(rx)) == rx


def test_complex_receiver_repr_roundtrip():
    rx = lox.ComplexReceiver(
        frequency=29e9 * lox.Hz,
        antenna_noise_temperature=265.0 * lox.K,
        lna_gain=30.0 * lox.dB,
        lna_noise_figure=1.0 * lox.dB,
        noise_figure=5.0 * lox.dB,
        loss=3.0 * lox.dB,
    )
    assert (
        eval(
            repr(rx),
            {
                "ComplexReceiver": lox.ComplexReceiver,
                "Frequency": lox.Frequency,
                "Temperature": lox.Temperature,
                "Decibel": lox.Decibel,
            },
        )
        == rx
    )


# --- Channel ---


def test_channel_bandwidth():
    # BPSK, 1 Mbit/s, roll-off=0.5, FEC=0.5 -> BW = 3 MHz
    ch = lox.Channel(
        link_type="downlink",
        data_rate=1e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("BPSK"),
        roll_off=0.5,
        fec=0.5,
    )
    assert ch.bandwidth().to_hertz() == pytest.approx(3e6, rel=1e-10)


def test_channel_eb_n0():
    # C/N0 = 80 dB*Hz, R = 1 Mbit/s -> Eb/N0 = 80 - 60 = 20 dB
    ch = lox.Channel(
        link_type="downlink",
        data_rate=1e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
    )
    eb_n0 = ch.eb_n0(80.0 * lox.dB)
    assert float(eb_n0) == pytest.approx(20.0, abs=1e-10)


def test_channel_link_margin():
    ch = lox.Channel(
        link_type="downlink",
        data_rate=1e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
    )
    margin = ch.link_margin(15.0 * lox.dB)
    assert float(margin) == pytest.approx(2.0, abs=1e-10)


def test_channel_invalid_link_type():
    with pytest.raises(ValueError, match="unknown link type"):
        lox.Channel(
            link_type="crosslink",
            data_rate=1e6 * lox.bps,
            required_eb_n0=10.0 * lox.dB,
            margin=3.0 * lox.dB,
            modulation=lox.Modulation("BPSK"),
        )


def test_channel_pickle():
    ch = lox.Channel(
        link_type="downlink",
        data_rate=1e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
        roll_off=0.35,
        fec=0.5,
    )
    restored = pickle.loads(pickle.dumps(ch))
    # Channel doesn't have __eq__, verify via bandwidth
    assert restored.bandwidth().to_hertz() == pytest.approx(
        ch.bandwidth().to_hertz(), abs=1e-10
    )


def test_channel_repr():
    ch = lox.Channel(
        link_type="downlink",
        data_rate=1e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
    )
    r = repr(ch)
    assert "link_type='downlink'" in r
    assert "modulation=Modulation('QPSK')" in r


# --- Environmental Losses ---


def test_environmental_losses_default():
    losses = lox.EnvironmentalLosses()
    assert float(losses.total()) == pytest.approx(0.0, abs=1e-15)


def test_environmental_losses_total():
    losses = lox.EnvironmentalLosses(
        rain=2.0 * lox.dB, gaseous=0.5 * lox.dB, atmospheric=1.0 * lox.dB
    )
    assert float(losses.total()) == pytest.approx(3.5, abs=1e-10)


def test_environmental_losses_eq():
    a = lox.EnvironmentalLosses(rain=2.0 * lox.dB)
    b = lox.EnvironmentalLosses(rain=2.0 * lox.dB)
    c = lox.EnvironmentalLosses(rain=3.0 * lox.dB)
    assert a == b
    assert not (a == c)


def test_environmental_losses_pickle():
    losses = lox.EnvironmentalLosses(
        rain=2.0 * lox.dB, gaseous=0.5 * lox.dB, atmospheric=1.0 * lox.dB
    )
    assert pickle.loads(pickle.dumps(losses)) == losses


def test_environmental_losses_repr_roundtrip():
    losses = lox.EnvironmentalLosses(
        rain=2.0 * lox.dB, gaseous=0.5 * lox.dB, atmospheric=1.0 * lox.dB
    )
    assert (
        eval(
            repr(losses),
            {"EnvironmentalLosses": lox.EnvironmentalLosses, "Decibel": lox.Decibel},
        )
        == losses
    )


# --- Free functions ---


def test_fspl():
    # 1000 km, 29 GHz -> ~181.7 dB
    loss = lox.fspl(distance=1000.0 * lox.km, frequency=29e9 * lox.Hz)
    assert float(loss) == pytest.approx(181.696, abs=0.1)


def test_freq_overlap_full():
    assert lox.freq_overlap(
        10e9 * lox.Hz, 1e6 * lox.Hz, 10e9 * lox.Hz, 1e6 * lox.Hz
    ) == pytest.approx(1.0, abs=1e-10)


def test_freq_overlap_none():
    assert lox.freq_overlap(
        10e9 * lox.Hz, 1e6 * lox.Hz, 12e9 * lox.Hz, 1e6 * lox.Hz
    ) == pytest.approx(0.0, abs=1e-10)


def test_freq_overlap_partial():
    assert lox.freq_overlap(
        10e9 * lox.Hz, 1e9 * lox.Hz, 10.5e9 * lox.Hz, 1e9 * lox.Hz
    ) == pytest.approx(0.5, abs=1e-10)


# --- Communication System ---


def test_communication_system_c_n0():
    tx_ant = lox.SimpleAntenna(gain=46.0 * lox.dB, beamwidth=0.7 * lox.deg)
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    tx_sys = lox.CommunicationSystem(antenna=tx_ant, transmitter=tx)

    rx_ant = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    rx_sys = lox.CommunicationSystem(antenna=rx_ant, receiver=rx)

    c_n0 = tx_sys.carrier_to_noise_density(
        rx_system=rx_sys,
        losses=0.0 * lox.dB,
        range=1000.0 * lox.km,
        tx_angle=0.0 * lox.deg,
        rx_angle=0.0 * lox.deg,
    )
    assert float(c_n0) == pytest.approx(104.9, abs=0.2)


def test_communication_system_noise_power():
    rx_ant = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    rx_sys = lox.CommunicationSystem(antenna=rx_ant, receiver=rx)

    p_noise = rx_sys.noise_power(bandwidth=1e6 * lox.Hz)
    assert float(p_noise) == pytest.approx(-141.61, abs=0.01)


def test_communication_system_pickle():
    tx_ant = lox.SimpleAntenna(gain=46.0 * lox.dB, beamwidth=0.7 * lox.deg)
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    tx_sys = lox.CommunicationSystem(antenna=tx_ant, transmitter=tx)
    restored = pickle.loads(pickle.dumps(tx_sys))
    # CommunicationSystem doesn't have __eq__, verify repr matches
    assert repr(restored) == repr(tx_sys)


def test_communication_system_pickle_with_receiver():
    rx_ant = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    rx_sys = lox.CommunicationSystem(antenna=rx_ant, receiver=rx)
    restored = pickle.loads(pickle.dumps(rx_sys))
    # Verify via noise_power computation
    assert float(restored.noise_power(1e6 * lox.Hz)) == pytest.approx(
        float(rx_sys.noise_power(1e6 * lox.Hz)), abs=1e-10
    )


# --- Link Stats ---


def test_link_stats_end_to_end():
    tx_ant = lox.SimpleAntenna(gain=46.0 * lox.dB, beamwidth=0.7 * lox.deg)
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    tx_sys = lox.CommunicationSystem(antenna=tx_ant, transmitter=tx)

    rx_ant = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    rx_sys = lox.CommunicationSystem(antenna=rx_ant, receiver=rx)

    ch = lox.Channel(
        link_type="downlink",
        data_rate=10e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
        roll_off=0.35,
        fec=0.5,
    )

    stats = lox.LinkStats.calculate(
        tx_system=tx_sys,
        rx_system=rx_sys,
        channel=ch,
        range=1000.0 * lox.km,
        tx_angle=0.0 * lox.deg,
        rx_angle=0.0 * lox.deg,
    )

    # EIRP = 46 + 10 - 1 = 55 dBW
    assert float(stats.eirp) == pytest.approx(55.0, abs=0.01)
    # FSPL at 1000 km, 29 GHz
    assert float(stats.fspl) == pytest.approx(181.7, abs=0.1)
    # C/N0
    assert float(stats.c_n0) == pytest.approx(104.9, abs=0.2)
    # Eb/N0 = C/N0 - 10*log10(10e6)
    assert float(stats.eb_n0) == pytest.approx(34.9, abs=0.2)
    # Margin = Eb/N0 - 10 - 3
    assert float(stats.margin) == pytest.approx(21.9, abs=0.2)
    # Properties
    assert stats.slant_range.to_kilometers() == pytest.approx(1000.0, abs=1e-6)
    assert stats.data_rate.to_bits_per_second() == pytest.approx(10e6, abs=1e-6)
    assert stats.frequency.to_hertz() == pytest.approx(29e9, abs=1.0)


def test_link_stats_with_losses():
    tx_ant = lox.SimpleAntenna(gain=46.0 * lox.dB, beamwidth=0.7 * lox.deg)
    tx = lox.Transmitter(
        frequency=29e9 * lox.Hz, power=10.0 * lox.W, line_loss=1.0 * lox.dB
    )
    tx_sys = lox.CommunicationSystem(antenna=tx_ant, transmitter=tx)

    rx_ant = lox.SimpleAntenna(gain=30.0 * lox.dB, beamwidth=3.0 * lox.deg)
    rx = lox.SimpleReceiver(
        frequency=29e9 * lox.Hz, system_noise_temperature=500.0 * lox.K
    )
    rx_sys = lox.CommunicationSystem(antenna=rx_ant, receiver=rx)

    ch = lox.Channel(
        link_type="downlink",
        data_rate=10e6 * lox.bps,
        required_eb_n0=10.0 * lox.dB,
        margin=3.0 * lox.dB,
        modulation=lox.Modulation("QPSK"),
        roll_off=0.35,
        fec=0.5,
    )

    losses = lox.EnvironmentalLosses(rain=2.0 * lox.dB, atmospheric=1.0 * lox.dB)

    stats_no_loss = lox.LinkStats.calculate(
        tx_sys, rx_sys, ch, 1000 * lox.km, 0 * lox.deg, 0 * lox.deg
    )
    stats_loss = lox.LinkStats.calculate(
        tx_sys, rx_sys, ch, 1000 * lox.km, 0 * lox.deg, 0 * lox.deg, losses
    )

    # 3 dB of environmental losses should reduce margin by 3 dB
    margin_diff = float(stats_no_loss.margin) - float(stats_loss.margin)
    assert margin_diff == pytest.approx(3.0, abs=0.01)



# SPDX-FileCopyrightText: 2025 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import math

import pytest

import lox_space as lox

# Reference values from lox-orbits Rust tests.
SMA_KM = 7178.1363
EARTH_R_KM = 6378.1366
EXP_INC_RAD = math.radians(98.627)
EXP_RAAN_RAD = math.radians(350.5997)


@pytest.fixture
def epoch():
    return lox.UTC.from_iso("2020-02-18T18:44:37.550").to_scale("TDB")


def test_sso_from_semi_major_axis(epoch, provider):
    """SSO from semi-major axis: derived inclination and RAAN match reference."""
    sso = lox.Keplerian.sso(
        epoch,
        semi_major_axis=SMA_KM * lox.km,
        ltan=(13, 30),
        provider=provider,
    )

    assert float(sso.inclination()) == pytest.approx(EXP_INC_RAD, rel=1e-5)
    assert float(sso.longitude_of_ascending_node()) == pytest.approx(
        EXP_RAAN_RAD, rel=4e-3
    )


def test_sso_from_altitude(epoch, provider):
    """SSO from altitude with LTDN gives the same result."""
    altitude = (SMA_KM - EARTH_R_KM) * lox.km

    sso = lox.Keplerian.sso(
        epoch,
        altitude=altitude,
        ltdn=(1, 30),
        provider=provider,
    )

    assert float(sso.inclination()) == pytest.approx(EXP_INC_RAD, rel=1e-5)
    assert float(sso.longitude_of_ascending_node()) == pytest.approx(
        EXP_RAAN_RAD, rel=4e-3
    )


def test_sso_invalid_parameters(epoch, provider):
    """Passing none or multiple of altitude/semi_major_axis/inclination raises."""
    with pytest.raises(ValueError, match="exactly one"):
        lox.Keplerian.sso(epoch, provider=provider)

    with pytest.raises(ValueError, match="exactly one"):
        lox.Keplerian.sso(
            epoch,
            altitude=800 * lox.km,
            semi_major_axis=7178 * lox.km,
            provider=provider,
        )


def test_sso_ltan_ltdn_exclusive(epoch, provider):
    """Passing both ltan and ltdn raises."""
    with pytest.raises(ValueError, match="at most one"):
        lox.Keplerian.sso(
            epoch,
            altitude=800 * lox.km,
            ltan=(13, 30),
            ltdn=(1, 30),
            provider=provider,
        )



# SPDX-FileCopyrightText: 2024 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import lox_space as lox
import numpy.testing as npt
import pytest


def test_state_to_ground_location():
    time = lox.UTC.from_iso("2024-07-05T09:09:18.173").to_scale("TAI")
    state = lox.Cartesian(
        time,
        position=[-5530017.74359, -3487089.5338, -1850034.76185],
        velocity=[1295.34407, -5024.56882, 5639.1936],
        origin=lox.Origin("Earth"),
        frame=lox.Frame("ICRF"),
    ).to_frame(lox.Frame("IAU_EARTH"))
    npt.assert_allclose(
        state.position() * 1e-3,
        [-5740.259426667957, 3121.1360727954725, -1863.1826563318027],
    )
    npt.assert_allclose(
        state.velocity() * 1e-3,
        [-3.53237875783652, -3.152377656863808, 5.642296713889555],
    )
    ground = state.to_ground_location()
    assert float(ground.longitude()) == pytest.approx(2.643578045424445)
    assert float(ground.latitude()) == pytest.approx(-0.27944957125091063)
    assert ground.altitude().to_kilometers() == pytest.approx(417.8524151150059)


def test_state_to_origin(ephemeris):
    r_m = [6068279.27, -1692843.94, -2516619.18]
    v_ms = [-660.415582, 5495.938726, -5303.093233]

    utc = lox.UTC.from_iso("2016-05-30T12:00:00.000")
    tai = utc.to_scale("TAI")

    s_earth = lox.Cartesian(
        tai,
        position=r_m,
        velocity=v_ms,
    )
    s_venus = s_earth.to_origin(lox.Origin("Venus"), ephemeris)

    # Verify round-trip: converting back to Earth origin should recover the original state
    s_earth_rt = s_venus.to_origin(lox.Origin("Earth"), ephemeris)

    npt.assert_allclose(s_earth_rt.position(), r_m, atol=1e-3)
    npt.assert_allclose(s_earth_rt.velocity(), v_ms, atol=1e-3)


def test_cartesian_component_kwargs():
    """Test Cartesian construction with x, y, z, vx, vy, vz keyword arguments."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        x=7000 * lox.km,
        y=0 * lox.km,
        z=0 * lox.km,
        vx=0 * lox.m_per_s,
        vy=7500 * lox.m_per_s,
        vz=0 * lox.m_per_s,
    )
    assert state.x.to_kilometers() == pytest.approx(7000.0)
    assert state.y.to_kilometers() == pytest.approx(0.0)
    assert state.z.to_kilometers() == pytest.approx(0.0)
    assert state.vx.to_meters_per_second() == pytest.approx(0.0)
    assert state.vy.to_meters_per_second() == pytest.approx(7500.0)
    assert state.vz.to_meters_per_second() == pytest.approx(0.0)


def test_cartesian_component_getters():
    """Test that x, y, z, vx, vy, vz return unit types."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 1000e3, 500e3],
        velocity=[100.0, 7500.0, -200.0],
    )
    assert isinstance(state.x, lox.Distance)
    assert isinstance(state.y, lox.Distance)
    assert isinstance(state.z, lox.Distance)
    assert isinstance(state.vx, lox.Velocity)
    assert isinstance(state.vy, lox.Velocity)
    assert isinstance(state.vz, lox.Velocity)
    assert state.x.to_meters() == pytest.approx(7000e3)
    assert state.vy.to_meters_per_second() == pytest.approx(7500.0)


def test_cartesian_repr():
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
    )
    r = repr(state)
    assert r.startswith("Cartesian(")
    assert "7000000.0" in r


def test_cartesian_string_origin_and_frame():
    """Test that string origin and frame work in Cartesian constructor."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
        origin="Earth",
        frame="ICRF",
    )
    assert state.origin().name() == "Earth"
    assert repr(state.reference_frame()) == 'Frame("ICRF")'


def test_cartesian_to_frame_string():
    """Test that to_frame accepts a string frame name."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
    )
    s2 = state.to_frame("ICRF")
    assert repr(s2.reference_frame()) == 'Frame("ICRF")'


def test_cartesian_to_origin_string(ephemeris):
    """Test that to_origin accepts a string origin name."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
    )
    s2 = state.to_origin("Moon", ephemeris)
    assert s2.origin().name() == "Moon"


# --- Keplerian constructor tests ---


def test_keplerian_positional_backward_compat():
    """Test that existing positional usage still works."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        7178.0 * lox.km,
        0.001,
        97.0 * lox.deg,
        0.0 * lox.deg,
        0.0 * lox.deg,
        0.0 * lox.deg,
    )
    assert k.semi_major_axis().to_kilometers() == pytest.approx(7178.0, rel=1e-10)
    assert k.eccentricity() == pytest.approx(0.001, abs=1e-15)


def test_keplerian_keyword_backward_compat():
    """Test that existing keyword usage still works."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        semi_major_axis=7178.0 * lox.km,
        eccentricity=0.001,
        inclination=97.0 * lox.deg,
        longitude_of_ascending_node=0.0 * lox.deg,
        argument_of_periapsis=0.0 * lox.deg,
        true_anomaly=0.0 * lox.deg,
    )
    assert k.semi_major_axis().to_kilometers() == pytest.approx(7178.0, rel=1e-10)
    assert k.eccentricity() == pytest.approx(0.001, abs=1e-15)
    assert k.inclination().to_degrees() == pytest.approx(97.0, rel=1e-10)


def test_keplerian_radii():
    """Test Keplerian construction from periapsis/apoapsis radii."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        periapsis_radius=7000.0 * lox.km,
        apoapsis_radius=7400.0 * lox.km,
    )
    assert k.semi_major_axis().to_kilometers() == pytest.approx(7200.0, rel=1e-10)
    exp_ecc = (7400.0 - 7000.0) / (7400.0 + 7000.0)
    assert k.eccentricity() == pytest.approx(exp_ecc, rel=1e-10)


def test_keplerian_altitudes():
    """Test Keplerian construction from periapsis/apoapsis altitudes."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        periapsis_altitude=600.0 * lox.km,
        apoapsis_altitude=1000.0 * lox.km,
    )
    # Mean radius of Earth ~6371 km
    sma_km = k.semi_major_axis().to_kilometers()
    assert sma_km > 6971.0
    assert sma_km < 7371.0
    assert k.eccentricity() > 0.0


def test_keplerian_defaults():
    """Test that angular elements default to 0."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        periapsis_radius=7000.0 * lox.km,
        apoapsis_radius=7000.0 * lox.km,
    )
    assert float(k.inclination()) == pytest.approx(0.0, abs=1e-15)
    assert float(k.longitude_of_ascending_node()) == pytest.approx(0.0, abs=1e-15)
    assert float(k.argument_of_periapsis()) == pytest.approx(0.0, abs=1e-15)
    assert float(k.true_anomaly()) == pytest.approx(0.0, abs=1e-15)


def test_keplerian_mean_anomaly():
    """Test Keplerian construction with mean_anomaly keyword."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian(
        time,
        7178.0 * lox.km,
        0.001,
        mean_anomaly=90.0 * lox.deg,
    )
    # mean anomaly of 90 deg should produce a non-zero true anomaly
    assert float(k.true_anomaly()) != pytest.approx(0.0, abs=1e-3)


def test_keplerian_both_anomalies_raises():
    """Test that providing both true_anomaly and mean_anomaly raises."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="true anomaly.*mean anomaly"):
        lox.Keplerian(
            time,
            7178.0 * lox.km,
            0.001,
            true_anomaly=0.0 * lox.deg,
            mean_anomaly=0.0 * lox.deg,
        )


def test_keplerian_no_shape_raises():
    """Test that omitting shape params raises ValueError."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="orbital shape"):
        lox.Keplerian(time)


def test_keplerian_mixed_shape_raises():
    """Test that mixing shape params raises ValueError."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="exactly one"):
        lox.Keplerian(
            time,
            7000.0 * lox.km,
            0.0,
            periapsis_radius=7000.0 * lox.km,
            apoapsis_radius=7000.0 * lox.km,
        )


def test_keplerian_partial_radii_raises():
    """Test that providing only one radius raises ValueError."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="periapsis_radius"):
        lox.Keplerian(time, apoapsis_radius=7000.0 * lox.km)


# --- Keplerian.circular() tests ---


def test_circular_from_sma():
    """Test circular orbit from semi-major axis."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian.circular(time, semi_major_axis=7178.0 * lox.km)
    assert k.semi_major_axis().to_kilometers() == pytest.approx(7178.0, rel=1e-10)
    assert k.eccentricity() == pytest.approx(0.0, abs=1e-15)


def test_circular_from_altitude():
    """Test circular orbit from altitude."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian.circular(time, altitude=800.0 * lox.km)
    # sma should be altitude + mean radius
    sma_km = k.semi_major_axis().to_kilometers()
    assert sma_km > 7100.0
    assert sma_km < 7200.0
    assert k.eccentricity() == pytest.approx(0.0, abs=1e-15)


def test_circular_with_inclination():
    """Test circular orbit with inclination."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    k = lox.Keplerian.circular(
        time,
        semi_major_axis=7178.0 * lox.km,
        inclination=97.0 * lox.deg,
    )
    assert k.inclination().to_degrees() == pytest.approx(97.0, rel=1e-10)


def test_circular_no_size_raises():
    """Test that omitting size raises ValueError."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="exactly one"):
        lox.Keplerian.circular(time)


def test_circular_both_size_raises():
    """Test that providing both sma and altitude raises ValueError."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    with pytest.raises(ValueError, match="exactly one"):
        lox.Keplerian.circular(
            time,
            semi_major_axis=7178.0 * lox.km,
            altitude=800.0 * lox.km,
        )



# SPDX-FileCopyrightText: 2024 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import lox_space as lox
import pytest


def test_time(provider):
    tai_exp = lox.Time("TAI", 2000, 1, 1)
    tai_act = lox.Time.from_iso("2000-01-01T00:00:00.000 TAI")
    assert tai_exp == tai_act
    tai_act = tai_exp.to_scale("TAI")
    assert tai_exp == tai_act
    tai_act = tai_exp.to_scale("TCB").to_scale("TAI")
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_scale("TCG").to_scale("TAI")
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_scale("TDB").to_scale("TAI")
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_scale("TT").to_scale("TAI")
    assert tai_exp.isclose(tai_act)
    tai_act = tai_exp.to_scale("UT1", provider).to_scale("TAI", provider)
    assert tai_exp.isclose(tai_act)
    tai1 = lox.Time("TAI", 2000, 1, 1, 0, 0, 0.5)
    assert tai1 > tai_exp
    assert tai1 >= tai_exp
    assert tai_exp < tai1
    assert tai_exp <= tai1
    assert tai_exp != tai1
    dt = lox.TimeDelta(0.5)
    assert (tai_exp + dt).isclose(tai1)
    assert (tai1 - dt).isclose(tai_exp)
    assert float(tai1 - tai_exp) == pytest.approx(float(dt))


def test_utc(provider):
    utc_exp = lox.UTC(2000, 1, 1)
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000")
    assert utc_exp == utc_act
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000Z")
    assert utc_exp == utc_act
    utc_act = lox.UTC.from_iso("2000-01-01T00:00:00.000 UTC")
    assert utc_exp == utc_act
    utc_act = utc_exp.to_scale("TAI").to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_scale("TCB").to_utc()
    assert utc_exp.isclose(utc_act)
    utc_act = utc_exp.to_scale("TCG").to_utc()
    assert utc_exp.isclose(utc_act)
    utc_act = utc_exp.to_scale("TDB").to_utc()
    assert utc_exp.isclose(utc_act)
    utc_act = utc_exp.to_scale("TT").to_utc()
    assert utc_exp == utc_act
    utc_act = utc_exp.to_scale("UT1", provider).to_utc(provider)
    assert utc_exp.isclose(utc_act)


def test_time_delta():
    delta = lox.TimeDelta(1.5)
    assert str(delta) == "1.5 seconds"
    assert repr(delta) == "TimeDelta(1.5)"
    assert delta.seconds() == 1
    assert delta.subsecond() == 0.5
    assert str(delta + delta) == "3 seconds"
    assert str(delta - delta) == "0 seconds"
    assert str(-delta) == "-1.5 seconds"
    with pytest.raises(lox.NonFiniteTimeDeltaError):
        lox.TimeDelta(float("nan")).seconds()


def test_time_delta_constructors():
    td = lox.TimeDelta.from_seconds(123)
    assert td.to_decimal_seconds() == 123.0
    td = lox.TimeDelta.from_minutes(2.0)
    assert td.to_decimal_seconds() == 120.0
    td = lox.TimeDelta.from_hours(2.0)
    assert td.to_decimal_seconds() == 7200.0
    td = lox.TimeDelta.from_days(2.0)
    assert td.to_decimal_seconds() == 172800.0
    td = lox.TimeDelta.from_julian_years(2.0)
    assert td.to_decimal_seconds() == 63115200.0
    td = lox.TimeDelta.from_julian_centuries(2.0)
    assert td.to_decimal_seconds() == 6311520000.0


def test_time_repr():
    time = lox.Time("TAI", 2000, 1, 1, 0, 0, 12.123456789123)
    assert repr(time) == 'Time("TAI", 2000, 1, 1, 0, 0, 12.123456789123)'
    assert str(time) == "2000-01-01T00:00:12.123 TAI"


def test_time_accessors():
    time = lox.Time("TAI", 2000, 1, 1, 0, 0, 12.123456789123)
    assert time.scale().abbreviation() == "TAI"
    assert time.year() == 2000
    assert time.month() == 1
    assert time.day() == 1
    assert time.hour() == 0
    assert time.minute() == 0
    assert time.second() == 12
    assert time.millisecond() == 123
    assert time.microsecond() == 456
    assert time.nanosecond() == 789
    assert time.picosecond() == 123
    assert time.femtosecond() == 0
    assert time.decimal_seconds() == pytest.approx(12.123456789123, rel=1e-15)


def test_time_invalid_date():
    with pytest.raises(ValueError, match="invalid date"):
        lox.Time("TAI", 2000, 13, 1)


def test_time_invalid_hour():
    with pytest.raises(ValueError, match="hour must be in the range"):
        lox.Time("TAI", 2000, 12, 1, 24, 0, 0.0)


def test_time_sub_different_scales():
    t1 = lox.Time("TAI", 2000, 1, 1, 0, 0, 1.0)
    t0 = lox.Time("TT", 2000, 1, 1, 0, 0, 1.0)
    with pytest.raises(ValueError, match="cannot subtract.*different time scales"):
        t1 - t0


def test_time_isclose_different_scales():
    t0 = lox.Time("TAI", 2000, 1, 1)
    t1 = lox.Time("TT", 2000, 1, 1)
    with pytest.raises(ValueError, match="cannot compare.*different time scales"):
        t0.isclose(t1)


def test_time_from_iso_invalid():
    with pytest.raises(ValueError, match="invalid ISO"):
        lox.Time.from_iso("2000-01-01X00:00:00 TAI")


def test_time_from_iso_invalid_scale():
    with pytest.raises(ValueError, match="invalid ISO"):
        lox.Time.from_iso("2000-01-01T00:00:00 UTC")


def test_time_from_iso_invalid_scale_arg():
    with pytest.raises(ValueError, match="unknown time scale: UTC"):
        lox.Time.from_iso("2000-01-01T00:00:00 TAI", scale="UTC")


def test_time_julian_date():
    time = lox.Time.from_julian_date("TAI", 0.0, "j2000")
    assert time.julian_date("j2000", "seconds") == 0.0
    assert time.julian_date("j2000", "days") == 0.0
    assert time.julian_date("j2000", "centuries") == 0.0
    assert time.julian_date("jd", "days") == 2451545.0
    assert time.julian_date("mjd", "days") == 51544.5
    assert time.julian_date("j1950", "days") == 18262.5

    time = lox.Time.from_julian_date("TAI", 0.0, "j1950")
    assert time.julian_date("j1950", "days") == 0.0

    time = lox.Time.from_julian_date("TAI", 0.0, "mjd")
    assert time.julian_date("mjd", "days") == 0.0

    time = lox.Time.from_julian_date("TAI", 0.0, "jd")
    assert time.julian_date("jd", "days") == 0.0


def test_time_invalid_epoch():
    time = lox.Time("TAI", 2000, 1, 1)
    with pytest.raises(ValueError, match="unknown epoch: unknown"):
        time.julian_date("unknown", "days")


def test_time_invalid_unit():
    time = lox.Time("TAI", 2000, 1, 1)
    with pytest.raises(ValueError, match="unknown unit: unknown"):
        time.julian_date("jd", "unknown")


def test_time_from_two_part_julian_date():
    expected = lox.Time("TAI", 2024, 7, 11, 8, 2, 14.0)
    jd1, jd2 = expected.two_part_julian_date()
    actual = lox.Time.from_two_part_julian_date("TAI", jd1, jd2)
    assert expected.isclose(actual)


def test_time_from_day_of_year():
    expected = lox.Time("TAI", 2024, 12, 31)
    actual = lox.Time.from_day_of_year("TAI", 2024, 366)
    assert actual == expected


def test_utc_accessors():
    utc = lox.UTC(2000, 1, 1, 12, 13, 14.123456789123)
    assert utc.year() == 2000
    assert utc.month() == 1
    assert utc.day() == 1
    assert utc.hour() == 12
    assert utc.minute() == 13
    assert utc.second() == 14
    assert utc.millisecond() == 123
    assert utc.microsecond() == 456
    assert utc.nanosecond() == 789
    assert utc.picosecond() == 123
    assert utc.decimal_seconds() == 14.123456789123
    assert str(utc) == "2000-01-01T12:13:14.123 UTC"
    assert repr(utc) == "UTC(2000, 1, 1, 12, 13, 14.123456789123)"


def test_utc_invalid_date():
    with pytest.raises(ValueError, match="invalid date"):
        lox.UTC(2000, 0, 1)


def test_utc_from_iso_invalid():
    with pytest.raises(ValueError, match="invalid ISO"):
        lox.UTC.from_iso("2000-01-01X00:00:00 UTC")


def test_time_delta_units():
    assert (60 * lox.seconds).to_decimal_seconds() == 60.0
    assert (2 * lox.minutes).to_decimal_seconds() == 120.0
    assert (0.5 * lox.hours).to_decimal_seconds() == 1800.0
    assert (1 * lox.days).to_decimal_seconds() == 86400.0


def test_time_delta_subsecond_constructors():
    td = lox.TimeDelta.from_milliseconds(1500)
    assert td.seconds() == 1
    assert td.subsecond() == 0.5
    td = lox.TimeDelta.from_microseconds(1_000_000)
    assert td.to_decimal_seconds() == 1.0
    td = lox.TimeDelta.from_nanoseconds(500_000_000)
    assert td.to_decimal_seconds() == 0.5


def test_eop_provider_invalid_path():
    with pytest.raises(lox.EopParserError):
        lox.EOPProvider("invalid_path")


def test_eop_provider_extrapolated(provider):
    tai = lox.Time("TAI", 2100, 1, 1)
    with pytest.raises(lox.EopProviderError, match="extrapolated"):
        tai.to_scale("UT1", provider)



# SPDX-FileCopyrightText: 2024 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import pathlib
from pathlib import Path

import pytest
import lox_space as lox

DATA_DIR = pathlib.Path(__file__).parents[3].joinpath("data")


@pytest.fixture(scope="session")
def data_dir():
    return DATA_DIR


@pytest.fixture(scope="session")
def provider():
    return lox.EOPProvider(
        DATA_DIR.joinpath("iers/finals.all.csv"),
        DATA_DIR.joinpath("iers/finals2000A.all.csv"),
    )


@pytest.fixture(scope="session")
def oneweb():
    with open(DATA_DIR.joinpath("oneweb_tle.txt")) as f:
        lines = f.readlines()

    sgp4s = []
    for i in range(0, len(lines), 3):
        tle = lines[i : i + 3]
        name = tle[0].strip()
        sgp4s.append((name, lox.SGP4("".join(tle))))

    return dict(sgp4s)


@pytest.fixture(scope="session")
def estrack():
    stations = [
        ("Kiruna", 67.858428, 20.966880),
        ("Esrange Space Center", 67.8833, 21.1833),
        ("Kourou", 5.2360, -52.7686),
        ("Redu", 50.00205516, 5.14518047),
        ("Cebreros", 40.3726, -4.4739),
        ("New Norcia", -30.9855, 116.2041),
    ]
    return [
        lox.GroundStation(
            name,
            lox.GroundLocation(
                lox.Origin("Earth"), lon * lox.deg, lat * lox.deg, 0 * lox.km
            ),
            lox.ElevationMask.fixed(0 * lox.rad),
        )
        for name, lat, lon in stations
    ]


@pytest.fixture(scope="session")
def ephemeris():
    return lox.SPK(DATA_DIR.joinpath("spice/de440s.bsp"))



# SPDX-FileCopyrightText: 2025 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import pytest
import lox_space as lox


def test_origin_from_string():
    earth = lox.Origin("Earth")
    assert earth.name() == "Earth"


def test_origin_from_id():
    earth = lox.Origin(399)
    assert earth.name() == "Earth"


def test_origin_repr():
    earth = lox.Origin("Earth")
    assert repr(earth) == 'Origin("Earth")'


def test_gravitational_parameter():
    earth = lox.Origin("Earth")
    gm = earth.gravitational_parameter()
    assert isinstance(gm, lox.GravitationalParameter)
    assert gm.to_km3_per_s2() == pytest.approx(398600.4418, rel=1e-4)


def test_mean_radius():
    earth = lox.Origin("Earth")
    r = earth.mean_radius()
    assert isinstance(r, lox.Distance)
    assert r.to_kilometers() == pytest.approx(6371.0084, rel=1e-3)


def test_radii():
    earth = lox.Origin("Earth")
    a, b, c = earth.radii()
    assert isinstance(a, lox.Distance)
    assert isinstance(b, lox.Distance)
    assert isinstance(c, lox.Distance)
    assert a.to_kilometers() == pytest.approx(6378.1366, rel=1e-3)


def test_equatorial_radius():
    earth = lox.Origin("Earth")
    r = earth.equatorial_radius()
    assert isinstance(r, lox.Distance)
    assert r.to_kilometers() == pytest.approx(6378.1366, rel=1e-3)


def test_polar_radius():
    earth = lox.Origin("Earth")
    r = earth.polar_radius()
    assert isinstance(r, lox.Distance)
    assert r.to_kilometers() == pytest.approx(6356.7519, rel=1e-3)


def test_rotational_elements():
    earth = lox.Origin("Earth")
    ra, dec, rot = earth.rotational_elements(0.0)
    assert isinstance(ra, lox.Angle)
    assert isinstance(dec, lox.Angle)
    assert isinstance(rot, lox.Angle)


def test_rotational_element_rates():
    earth = lox.Origin("Earth")
    ra_rate, dec_rate, rot_rate = earth.rotational_element_rates(0.0)
    assert isinstance(ra_rate, lox.AngularRate)
    assert isinstance(dec_rate, lox.AngularRate)
    assert isinstance(rot_rate, lox.AngularRate)


def test_right_ascension():
    earth = lox.Origin("Earth")
    ra = earth.right_ascension(0.0)
    assert isinstance(ra, lox.Angle)


def test_right_ascension_rate():
    earth = lox.Origin("Earth")
    ra_rate = earth.right_ascension_rate(0.0)
    assert isinstance(ra_rate, lox.AngularRate)


def test_declination():
    earth = lox.Origin("Earth")
    dec = earth.declination(0.0)
    assert isinstance(dec, lox.Angle)


def test_declination_rate():
    earth = lox.Origin("Earth")
    dec_rate = earth.declination_rate(0.0)
    assert isinstance(dec_rate, lox.AngularRate)


def test_rotation_angle():
    earth = lox.Origin("Earth")
    rot = earth.rotation_angle(0.0)
    assert isinstance(rot, lox.Angle)


def test_rotation_rate():
    earth = lox.Origin("Earth")
    rot_rate = earth.rotation_rate(0.0)
    assert isinstance(rot_rate, lox.AngularRate)


def test_origin_string_in_cartesian():
    """Test that string origin works in Cartesian constructor."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
        origin="Earth",
    )
    assert state.origin().name() == "Earth"


def test_origin_id_in_cartesian():
    """Test that integer origin ID works in Cartesian constructor."""
    time = lox.UTC(2024, 1, 1).to_scale("TDB")
    state = lox.Cartesian(
        time,
        position=[7000e3, 0.0, 0.0],
        velocity=[0.0, 7500.0, 0.0],
        origin=399,
    )
    assert state.origin().name() == "Earth"



# SPDX-FileCopyrightText: 2025 Helge Eichhorn <git@helgeeichhorn.de>
#
# SPDX-License-Identifier: MPL-2.0

import math
import pytest
import lox_space as lox


def test_angle_rad():
    a = math.pi * lox.rad
    assert str(a) == "180 deg"
    assert repr(a) == "Angle(3.141592653589793)"
    assert float(a) == math.pi
    assert int(a) == 3
    assert complex(a) == complex(math.pi)


def test_angle_deg():
    a = 180 * lox.deg
    assert str(a) == "180 deg"
    assert repr(a) == "Angle(3.141592653589793)"
    assert float(a) == math.pi
    assert int(a) == 3
    assert complex(a) == complex(math.pi)


def test_distance_km():
    d = 1024 * lox.km
    assert str(d) == "1024 km"
    assert repr(d) == "Distance(1024000.0)"
    assert float(d) == 1024000.0
    assert int(d) == 1024000
    assert complex(d) == complex(1024000.0)


def test_distance_m():
    d = 2048 * lox.m
    assert str(d) == "2.048 km"
    assert repr(d) == "Distance(2048.0)"
    assert float(d) == 2048.0
    assert int(d) == 2048
    assert complex(d) == complex(2048.0)


def test_frequency_hz():
    f = 1073741824 * lox.Hz
    assert str(f) == "1.073741824 GHz"
    assert repr(f) == "Frequency(1073741824.0)"
    assert float(f) == 1073741824.0
    assert int(f) == 1073741824
    assert complex(f) == complex(1073741824.0)


def test_frequency_khz():
    f = 2000000 * lox.kHz
    assert str(f) == "2 GHz"
    assert repr(f) == "Frequency(2000000000.0)"
    assert float(f) == 2000000000.0
    assert int(f) == 2000000000
    assert complex(f) == complex(2000000000.0)


def test_velocity_ms():
    v = 262144 * lox.m_per_s
    assert str(v) == "262.144 km/s"
    assert repr(v) == "Velocity(262144.0)"
    assert float(v) == 262144.0
    assert int(v) == 262144
    assert complex(v) == complex(262144.0)


def test_velocity_kms():
    v = 16 * lox.km_per_s
    assert str(v) == "16 km/s"
    assert repr(v) == "Velocity(16000.0)"
    assert float(v) == 16000.0
    assert int(v) == 16000
    assert complex(v) == complex(16000.0)


# --- Arithmetic operations (covers __add__, __sub__, __neg__, __mul__, __eq__) ---


def test_distance_arithmetic():
    a = 100 * lox.m
    b = 200 * lox.m
    assert a + b == lox.Distance(300.0)
    assert b - a == lox.Distance(100.0)
    assert -a == lox.Distance(-100.0)
    assert a * 3 == lox.Distance(300.0)


def test_angle_arithmetic():
    a = 1.0 * lox.rad
    b = 2.0 * lox.rad
    assert a + b == lox.Angle(3.0)
    assert b - a == lox.Angle(1.0)
    assert -a == lox.Angle(-1.0)
    assert a * 2 == lox.Angle(2.0)


def test_velocity_arithmetic():
    a = 10 * lox.m_per_s
    b = 20 * lox.m_per_s
    assert a + b == lox.Velocity(30.0)
    assert b - a == lox.Velocity(10.0)
    assert -a == lox.Velocity(-10.0)


def test_frequency_eq():
    assert 1000 * lox.Hz == 1 * lox.kHz


# --- Conversion methods ---


def test_angle_conversions():
    a = math.pi * lox.rad
    assert a.to_radians() == pytest.approx(math.pi)
    assert a.to_degrees() == pytest.approx(180.0)
    a2 = lox.Angle(math.pi / 180.0 / 3600.0)
    assert a2.to_arcseconds() == pytest.approx(1.0)


def test_distance_conversions():
    d = 1 * lox.km
    assert d.to_meters() == pytest.approx(1000.0)
    assert d.to_kilometers() == pytest.approx(1.0)
    au = lox.Distance(lox.au.to_meters())
    assert au.to_astronomical_units() == pytest.approx(1.0)


def test_velocity_conversions():
    v = 1 * lox.km_per_s
    assert v.to_meters_per_second() == pytest.approx(1000.0)
    assert v.to_kilometers_per_second() == pytest.approx(1.0)


def test_frequency_conversions():
    f = 1 * lox.GHz
    assert f.to_hertz() == pytest.approx(1e9)
    assert f.to_kilohertz() == pytest.approx(1e6)
    assert f.to_megahertz() == pytest.approx(1e3)
    assert f.to_gigahertz() == pytest.approx(1.0)
    assert f.to_terahertz() == pytest.approx(1e-3)


# --- New unit types ---


def test_angular_rate():
    ar = 1.0 * lox.rad_per_s
    assert str(ar)
    assert repr(ar) == "AngularRate(1.0)"
    assert float(ar) == 1.0
    assert ar.to_radians_per_second() == pytest.approx(1.0)
    assert ar.to_degrees_per_second() == pytest.approx(math.degrees(1.0))


def test_angular_rate_deg_per_s():
    ar = 180.0 * lox.deg_per_s
    assert ar.to_radians_per_second() == pytest.approx(math.pi)
    assert ar.to_degrees_per_second() == pytest.approx(180.0)


def test_angular_rate_arithmetic():
    a = 1.0 * lox.rad_per_s
    b = 2.0 * lox.rad_per_s
    assert a + b == lox.AngularRate(3.0)
    assert b - a == lox.AngularRate(1.0)
    assert -a == lox.AngularRate(-1.0)


def test_data_rate():
    dr = 1000 * lox.bps
    assert repr(dr) == "DataRate(1000.0)"
    assert float(dr) == 1000.0
    assert dr.to_bits_per_second() == pytest.approx(1000.0)
    assert dr.to_kilobits_per_second() == pytest.approx(1.0)
    assert dr.to_megabits_per_second() == pytest.approx(0.001)


def test_data_rate_kbps():
    dr = 1 * lox.kbps
    assert dr.to_bits_per_second() == pytest.approx(1000.0)


def test_data_rate_mbps():
    dr = 1 * lox.Mbps
    assert dr.to_bits_per_second() == pytest.approx(1e6)


def test_data_rate_arithmetic():
    a = 100 * lox.bps
    b = 200 * lox.bps
    assert a + b == lox.DataRate(300.0)
    assert b - a == lox.DataRate(100.0)


def test_power():
    p = 100 * lox.W
    assert repr(p) == "Power(100.0)"
    assert float(p) == 100.0
    assert p.to_watts() == pytest.approx(100.0)
    assert p.to_kilowatts() == pytest.approx(0.1)
    assert p.to_dbw() == pytest.approx(20.0)


def test_power_kw():
    p = 1 * lox.kW
    assert p.to_watts() == pytest.approx(1000.0)


def test_power_arithmetic():
    a = 50 * lox.W
    b = 150 * lox.W
    assert a + b == lox.Power(200.0)
    assert b - a == lox.Power(100.0)
    assert -a == lox.Power(-50.0)


def test_temperature():
    t = 290 * lox.K
    assert repr(t) == "Temperature(290.0)"
    assert float(t) == 290.0
    assert t.to_kelvin() == pytest.approx(290.0)


def test_temperature_arithmetic():
    a = 100 * lox.K
    b = 200 * lox.K
    assert a + b == lox.Temperature(300.0)
    assert b - a == lox.Temperature(100.0)


# --- GravitationalParameter ---


def test_gravitational_parameter():
    gm = lox.GravitationalParameter(3.986004418e14)
    assert repr(gm) == "GravitationalParameter(398600441800000.0)"
    assert float(gm) == 3.986004418e14
    assert gm.to_m3_per_s2() == pytest.approx(3.986004418e14)
    assert gm.to_km3_per_s2() == pytest.approx(398600.4418)


def test_gravitational_parameter_from_km3():
    gm = lox.GravitationalParameter.from_km3_per_s2(398600.4418)
    assert gm.to_m3_per_s2() == pytest.approx(3.986004418e14)
    assert gm.to_km3_per_s2() == pytest.approx(398600.4418)


def test_gravitational_parameter_eq():
    gm1 = lox.GravitationalParameter(1e14)
    gm2 = lox.GravitationalParameter(1e14)
    assert gm1 == gm2


# --- __getnewargs__ (pickle support) ---


def test_distance_getnewargs():
    d = 42.0 * lox.m
    args = d.__getnewargs__()
    assert args == (42.0,)
    assert lox.Distance(*args) == d


def test_gravitational_parameter_getnewargs():
    gm = lox.GravitationalParameter(1e14)
    args = gm.__getnewargs__()
    assert lox.GravitationalParameter(*args) == gm


# --- repr with integer values (covers repr_f64 for non-decimal values) ---


def test_repr_integer_value():
    d = lox.Distance(1000.0)
    assert repr(d) == "Distance(1000.0)"


# --- Module-level dB constant ---


def test_db_constant():
    db = 3 * lox.dB
    assert float(db) == pytest.approx(3.0)


def test_decibel_mul():
    db = lox.Decibel(2.0)
    assert float(db * 3) == pytest.approx(6.0)
    assert float(3 * db) == pytest.approx(6.0)
